import torch 
import torchvision
import numpy as np
import math
import matplotlib.pyplot as plt

from tqdm import tqdm 

def get_mean_std(dataset_loader):
    mean = torch.zeros(3)
    std = torch.zeros(3)
    print('==> Computing mean and std..')
    for k in tqdm(dataset_loader):
        inputs, _ = k["img"], k["label"]
        for i in range(3):
            mean[i] += inputs[:,i,:,:].mean()
            std[i] += inputs[:,i,:,:].std()
    mean.div_(len(dataset_loader))
    std.div_(len(dataset_loader))
    print( mean, std)
    return mean,std

def imshow(images, labels, imgs_per_row=4):
    """
    images - stack of images 
    """
    opt = config()
    mean = list(opt.mean)
    std = list(opt.std)
    for img in images:
        for i in range(3):
            img[i] = img[i]*std[i] + mean[i]
    pil_convertor = torchvision.transforms.ToPILImage(mode='RGB')
    pil_images = [ pil_convertor(img) for img in images ]
    batches = math.ceil(len(pil_images)/float(imgs_per_row))
    for i in range(batches):
        imgs = pil_images[i*imgs_per_row:(i+1)*imgs_per_row]
        lab = labels[i*imgs_per_row:(i+1)*imgs_per_row]
        fig, ax = plt.subplots(nrows=1, ncols=len(imgs), sharex="col", sharey="row", figsize=(4*(len(imgs)),4), squeeze=False)
        for i, img in enumerate(imgs):    
            ax[0,i].imshow(img)
            ax[0,i].set_title(lab[i].item())

def vis_unnormalise(train_dataloader, nums):
    dataiter = iter(train_dataloader)
    images, labels = dataiter.next()
    imshow(images[:nums], labels[:nums])

    
def inference_imshow(images, pred_prob, imgs_per_row=3):
    """
    images - stack of images 
    pred_prob - contains list of probabilities and predicted class(incase of incorrect predictions)
    """
    print(type(images))
    print(images.shape)
    print(type(pred_prob))
    pil_convertor = torchvision.transforms.ToPILImage(mode='RGB')
    pil_images = [ pil_convertor(img) for img in images ]
    batches = math.ceil(len(pil_images)/float(imgs_per_row))
    for i in range(batches):
        imgs = pil_images[i*imgs_per_row:(i+1)*imgs_per_row]
        lab = pred_prob[i*imgs_per_row:(i+1)*imgs_per_row]
        fig, ax = plt.subplots(nrows=1, ncols=len(imgs), sharex="col", sharey="row", figsize=(5*(len(imgs)),4), squeeze=False)
        for i, img in enumerate(imgs):    
            ax[0,i].imshow(img)
            ax[0,i].set_title(lab[i])


if __name__ == '__main__':
    from .loaders import ListDataset
    from ..cfgs.ifood import config
    opt = config()
    opt.normalize = False
    trainset = ListDataset(opt, train=True)
    print(len(trainset))
    trainloader = torch.utils.data.DataLoader(trainset, opt.batch_size, shuffle=True, num_workers=4)
    mean, std = get_mean_std(trainloader)



import torch
import numpy as np
from PIL import Image
from tqdm import tqdm 
import os
import pickle

from torchvision import transforms

from .img_transforms import MyResizer, Normalizer

class ListDataset():
    def __init__(self,
                 opt,
                 train=True):
        self.opt = opt
        self.train = train
        self.multiclass = self.opt.multiclass
        #self.aspect_loc = self.opt.aspect_loc
        if self.multiclass:
            from sklearn.preprocessing import MultiLabelBinarizer
            self.one_hot = MultiLabelBinarizer(classes=np.arange(self.opt.num_classes))
        if self.train:
            self.list_file = self.opt.train_data_loc
        else:
            self.list_file = self.opt.val_data_loc
            
        
        self.input_size = self.opt.input_size
        self.data_loc = self.opt.data_loc
        
        
        if self.train:
            self.transforms = ListDataset.train_transformations(self.opt)
        else:
            self.transforms = ListDataset.test_transformations(self.opt)
        self.images, self.labels = [], []

        with open(self.list_file) as f:
            lines = f.readlines()
            self.num_samples = len(lines)

        for line in lines:
            splited = line.strip().rsplit(" ")
            self.images.append(splited[0])
            if self.multiclass:
                self.labels.append([int(i) for i in splited[1:]])
            else:
                self.labels.append(torch.LongTensor([int(splited[1])]))
        print("images:", len(self.images))
        
    def __getitem__(self, idx):
        img_loc = self.images[idx]
        label = self.labels[idx]
        img = self.get_image(img_loc)
        if self.multiclass:
            label = torch.FloatTensor(self.one_hot.fit_transform([label])).view(-1)
        else:
            label = label.long()
        return img, label, img_loc

    def __len__(self):
        return len(self.images)

    def get_image(self, img_loc):
        img = Image.open(self.data_loc+img_loc)
        if img.mode != self.opt.input_space:
            img = img.convert(self.opt.input_space)
        #img = np.asarray(img)
        img = self.transforms(img)
        return img

    def inverse_image(self):
        pass ## Given torch output tensor output of a dataloader, this should reverse engineer the options and output another image.
        
    @staticmethod
    def train_transformations(opt):
        img_transforms = []
        
        if opt.rotation:
            img_transforms.append(transforms.RandomRotation(2))
        if opt.h_flip:
            img_transforms.append(transforms.RandomHorizontalFlip())
        if opt.v_flip:
            img_transforms.append(transforms.RandomVerticalFlip())
        if opt.hue > 0 or opt.saturation > 0 or opt.contrast > 0 or opt.brightness >0:
            img_transforms.append(transforms.ColorJitter(brightness=opt.brightness,
                                             contrast=opt.contrast,
                                             saturation=opt.saturation,
                                             hue=opt.hue))
        if opt.pad:
            img_transforms.append(transforms.ToTensor())
            m = torch.nn.ReflectionPad2d((4,4,4,4))
            img_transforms.append(transforms.Lambda(lambda x: m(x.unsqueeze(0)).data.squeeze()))
            img_transforms.append(transforms.ToPILImage())
        if isinstance(opt.input_size, tuple):
            input_size = opt.input_size[0]
        else:
            input_size = opt.input_size
        if opt.resize:
            img_transforms.append(transforms.Resize((input_size, input_size)))
        img_transforms.append(transforms.ToTensor())
        
        if opt.normalize:
            img_transforms.append(transforms.Normalize(opt.mean, opt.std))
        
        return transforms.Compose(img_transforms)
    
    @staticmethod
    def test_transformations(opt):
        img_transforms = []
        
        if isinstance(opt.input_size, tuple):
            input_size = opt.input_size[0]
        else:
            input_size = opt.input_size
        
        if opt.use_tencrop:
            if opt.resize:
                img_transforms.append(transforms.Resize(input_size))
            img_transforms.append(transforms.TenCrop(input_size))
            img_transforms.append(transforms.Lambda(lambda crops: torch.stack([transforms.ToTensor()(crop) for crop in crops])))
            if opt.normalize:
                img_transforms.append(transforms.Lambda(lambda crops:torch.stack([transforms.Normalize(opt.mean, opt.std)(crop) for crop in crops])))
        else:
            if opt.resize:
                img_transforms.append(transforms.Resize((input_size, input_size)))
            img_transforms.append(transforms.ToTensor())
            if opt.normalize:
                img_transforms.append(transforms.Normalize(opt.mean, opt.std))
            
        return transforms.Compose(img_transforms)



import torch
from person_tracking.fractal.anchors.anchors_generator import Yolov3
# from fractal.anchors.anchors_generator import Yolov3

def get_anchors(model_name):
    if model_name == "yolov3":
        print("Using yolov3 model")
        return Yolov3
    else:
        raise NotImplementedError("We haven't implemented these anchor boxes yet. Please raise a pull request if it is useful")


""" Optimizers to train deep neural networks
"""
__author__ = ["PrakashJay"]

import torch

def get_optim(optim_type, model, kwargs):
    if optim_type == "sgd":
        print("Using yolov3 loss model")
        return sgd(optim_type, model, kwargs)
        
    else:
        raise NotImplementedError("We haven't implemented these anchor boxes yet. Please raise a pull request if it is useful")


def sgd(optim_type, model, kwargs):
    if kwargs is None:
        print("Using Default values for optimizer")
        kwargs = {"lr": 0.001, 
        "momentum": 0.9,
        "weight_decay": 0.0005,
        "nesterov": True}
    return torch.optim.SGD(model.parameters(),
                           lr = kwargs["lr"], 
                           momentum = kwargs["momentum"],
                           weight_decay= kwargs["weight_decay"],
                           nesterov=kwargs["nesterov"])


import torch 
import torchvision.transforms as transforms
from torch.utils.data.sampler import Sampler

import numpy as np
import skimage
import random

class RectangularCropTfm(object):
    """ Rectangular Image transforms as told in fastai library
    """
    def __init__(self, idx2ar, target_size):
        self.idx2ar, self.target_size = idx2ar, target_size

    def __call__(self, img):
        target_ar = self.idx2ar[idx]
        if target_ar < 1: 
            w = int(self.target_size/target_ar)
            size = (w//8*8, self.target_size)
        else: 
            h = int(self.target_size*target_ar)
            size = (self.target_size, h//8*8)
        return transforms.functional.center_crop(img, size)
    
    
class MyResizer(object):
    """Convert ndarrays in sample to Tensors."""
    def __init__(self, min_side=224, max_side=512):
        self.min_side = min_side
        self.max_side = max_side

    def __call__(self, image):
        image= image['image']

        rows, cols, cns = image.shape

        smallest_side = min(rows, cols)

        # rescale the image so the smallest side is min_side
        scale = self.min_side / smallest_side

        # check if the largest side is now greater than max_side, which can happen
        # when images have a large aspect ratio
        largest_side = max(rows, cols)

        if largest_side * scale > self.max_side:
            scale = max_side / largest_side

        # resize the image with the computed scale
        image = skimage.transform.resize(image, (int(round(rows*scale)), int(round((cols*scale)))),
                                        anti_aliasing=True, mode="constant")#, mode='constant', anti_aliasing=False)
        rows, cols, cns = image.shape

        pad_w = 32 - rows%32
        pad_h = 32 - cols%32

        new_image = np.zeros((rows + pad_w, cols + pad_h, cns)).astype(np.float32)
        new_image[:rows, :cols, :] = image.astype(np.float32)

        return {"image": torch.from_numpy(new_image), "scale":scale} 
    
class Normalizer(object):

    def __init__(self, mean = [0.485, 0.456, 0.406],
                std = [0.229, 0.224, 0.225]):
        self.mean = np.array([[mean]])
        self.std = np.array([[std]])

    def __call__(self, image):

        image = image['image']
        return {'image':((image.astype(np.float32)-self.mean)/self.std)}
    

class UnNormalizer(object):
    def __init__(self, mean=None, std=None):
        if mean == None:
            self.mean = [0.485, 0.456, 0.406]
        else:
            self.mean = mean
        if std == None:
            self.std = [0.229, 0.224, 0.225]
        else:
            self.std = std

    def __call__(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
        Returns:
            Tensor: Normalized image.
        """
        for t, m, s in zip(tensor, self.mean, self.std):
            t.mul_(s).add_(m)
        return tensor

def collater(data):

    imgs = [s['img'] for s in data]
    scales = [s['scale'] for s in data]
    labels = [s["label"] for s in data]
        
    widths = [int(s.shape[0]) for s in imgs]
    heights = [int(s.shape[1]) for s in imgs]
    batch_size = len(imgs)

    max_width = np.array(widths).max()
    max_height = np.array(heights).max()

    padded_imgs = torch.zeros(batch_size, max_width, max_height, 3)

    for i in range(batch_size):
        img = imgs[i]
        padded_imgs[i, :int(img.shape[0]), :int(img.shape[1]), :] = img
    

    padded_imgs = padded_imgs.permute(0, 3, 1, 2)
    labels = torch.cat([i.view(1, -1) for i in labels])

    return {'img': padded_imgs, 'scale': scales, "label": labels}


class AspectRatioBasedSampler(Sampler):

    def __init__(self, data_source, batch_size, drop_last):
        self.data_source = data_source
        self.batch_size = batch_size
        self.drop_last = drop_last
        self.groups = self.group_images()

    def __iter__(self):
        random.shuffle(self.groups)
        for group in self.groups:
            yield group

    def __len__(self):
        if self.drop_last:
            return len(self.data_source) // self.batch_size
        else:
            return (len(self.data_source) + self.batch_size - 1) // self.batch_size

    def group_images(self):
        # determine the order of the images
        order = list(range(len(self.data_source)))
        order.sort(key=lambda x: self.data_source.image_aspect_ratio(x))

        # divide into groups, one group = one batch
        return [[order[x % len(order)] for x in range(i, i + self.batch_size)] for i in range(0, len(order), self.batch_size)]


import torch
import torch.nn as nn

class ConvBlock(nn.Module):
    def __init__(self, inplanes, planes, kernel_size, stride=1, pad=1, bias=False):
        super(ConvBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=inplanes,
                             out_channels = planes,
                             kernel_size = kernel_size,
                             bias = bias,
                             stride = stride,
                             padding = pad)
        self.bn1 = nn.BatchNorm2d(planes)
        self.lrelu = nn.LeakyReLU(negative_slope=0.1, inplace=True)

    def forward(self, x):
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.lrelu(out)
        return out
    
    
class ResBlock(nn.Module):
    def __init__(self, in_channels, reduce):
        super(ResBlock, self).__init__()
        self.conv1 = ConvBlock(in_channels, reduce, 1, 1, 0)
        self.conv2 = ConvBlock(reduce, in_channels, 3, 1, 1)
    
    def forward(self, x):
        out = self.conv1(x)
        out = self.conv2(out)
        return out+x
        

class Darknet53(nn.Module):
    def __init__(self):
        super(Darknet53, self).__init__()
        self.conv1 = ConvBlock(3, 32, 3, 1, 1) # 256 * 256
        self.conv2 = ConvBlock(32, 64, 3, 2, 1) # 128 * 128
        self.resblock1 = self._make_layers(64, 32, 1)
        self.conv3 = ConvBlock(64, 128, 3, 2, 1)
        self.resblock2 = self._make_layers(128, 64, 2)
        self.conv4 = ConvBlock(128, 256, 3, 2, 1)
        self.resblock3 = self._make_layers(256, 128, 8)
        self.conv5 = ConvBlock(256, 512, 3, 2, 1)
        self.resblock4 = self._make_layers(512, 256, 8)
        self.conv6 = ConvBlock(512, 1024, 3, 2, 1)
        self.resblock5 = self._make_layers(1024, 512, 4)

    
    def _make_layers(self, in_channels, reduce, blocks):
        layer = []
        for _ in range(blocks):
            layer.append(ResBlock(in_channels, reduce))
        return nn.Sequential(*layer)
    
    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.resblock1(x)
        x = self.conv3(x)
        x = self.resblock2(x)
        x = self.conv4(x)
        out1 = self.resblock3(x)
        x = self.conv5(out1)
        out2 = self.resblock4(x)
        x = self.conv6(out2)
        out3 = self.resblock5(x)
        return out3
    

class Yolov3_classifier(nn.Module):
    def __init__(self, num_classes = 211):
        super(Yolov3_classifier, self).__init__()
        self.num_classes = num_classes
        self.backend = Darknet53()
        self.avgpool = nn.AvgPool2d(13, stride=1)
        self.classifier = nn.Linear(1024, self.num_classes)
    
    def forward(self, x):
        x = self.backend(x)
        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.classifier(x)
        return x 



from collections import defaultdict 

import numpy as np 
import six
from ..anchors.utils import bbox_iou

class Voc_mAP:
    """ Extracts the ground truth objects and prediction objects from respective files and align them on per image basis. 
    """
    def __init__(self, actual_file_loc, predicted_file_loc, labels_loc):
        self.actual_file_loc = actual_file_loc
        self.predicted_file_loc = predicted_file_loc
        self.labels_loc = labels_loc
        #self.type= type #image wise or category wise
        
        self._gts= self.load_datasets(self.actual_file_loc)
        self._dts= self.load_datasets(self.predicted_file_loc, "dt")
        
        self.pred_bboxes = list()
        self.pred_labels = list()
        self.pred_scores = list()
        self.gt_bboxes = list()
        self.gt_labels = list()
        
        
    def load_datasets(self, loc, gt_or_dt ="gt"):
        if isinstance(loc, str):
            with open(loc) as f:
                lines = f.readlines()
        elif isinstance(loc,list):
            lines=loc
        else:
            print ("issue in mAP code.")
    
        data = {}
        for line in lines:
            splited = line.strip().split(' ')
            if gt_or_dt == "dt":
                num_boxes = (len(splited) - 1)//6
            else:
                num_boxes = (len(splited) - 1)//5
            box = []
            label = []
            if gt_or_dt == "dt":
                score = []
            for i in range(num_boxes):
                if gt_or_dt == "dt":
                    ymin = splited[1+6*i]
                    xmin = splited[2+6*i]
                    ymax = splited[3+6*i]
                    xmax = splited[4+6*i]
                    c = int(splited[5+6*i])
                    s = splited[6+6*i]
                else:
                    ymin = splited[1+5*i]
                    xmin = splited[2+5*i]
                    ymax = splited[3+5*i]
                    xmax = splited[4+5*i]
                    c = int(splited[5+5*i]) 
                box.append([float(ymin), float(xmin), float(ymax), float(xmax)])
                label.append(c)
                if gt_or_dt == "dt":
                    score.append(float(s))
            if gt_or_dt == "dt":
                data[splited[0]] = [np.asarray(box), np.asarray(label), np.asarray(score)]
            else:
                data[splited[0]] = [np.asarray(box), np.asarray(label)]
        return data
    
    def cal_mAP(self, iou_thresh=0.5, use_07_metric=False):
        imgs_list = list(self._dts.keys())
        
        for i in imgs_list:
            if i not in self._dts.keys():
                raise ValueError('Not all the images in the test are predicted. Please check once again.')
        
        if len(self.pred_bboxes) == 0:
            for i in imgs_list:
                pb, pl, ps = self._dts[i]
                gb, gl = self._gts[i]
                gl = gl-1
                self.pred_bboxes.append(pb)
                self.pred_labels.append(pl)
                self.pred_scores.append(ps)
                self.gt_bboxes.append(gb)
                self.gt_labels.append(gl)
        
        return eval_detection_voc(self.pred_bboxes,
                                 self.pred_labels,
                                 self.pred_scores,
                                 self.gt_bboxes,
                                 self.gt_labels,
                                 iou_thresh=iou_thresh,
                                 use_07_metric=use_07_metric)
    
def eval_detection_voc(pred_bboxes, 
                       pred_labels, 
                       pred_scores, 
                       gt_bboxes, 
                       gt_labels, 
                       iou_thresh=0.5, 
                       use_07_metric=False):
    """ Calculate the Average precision of each object class and the mean Average precision of all the cass
    Args:
        pred_bboxes: A list of arrays, where each array represents the bboxes predicted for each image
        pred_labels: A list of arrays, where each array represents the corresponding labels for the predicted pred_bboxes 
        pred_scores: A list of arrays, where each array represents the corresponding probability scores for the predicted labels
        gt_boxes: A list of arrays, where each array represents the ground truth bboxes for each image 
        gt_labels: A list of arrays, where each array represents the corresponing labels for the ground truth bboxes 
        iou_thresh: minimum threshold required to consider this as an image 
        use_07_metric: Weather to use the evaluation metric discussed in the VOC evaluation paper [1]. Default false.

    Returns:
            A dict:  AP: average precision of each object 
                    mAP: mean average precision of all the objects 
    """

    prec, rec = calc_detection_voc_prec_rec(
    pred_bboxes, pred_labels, pred_scores,
    gt_bboxes, gt_labels,
    iou_thresh=iou_thresh)

    ap = calc_detection_voc_ap(prec, rec, use_07_metric=use_07_metric)

    return {'ap': ap, 'map': np.nanmean(ap)}

    
    

def calc_detection_voc_prec_rec(
        pred_bboxes, pred_labels, pred_scores, gt_bboxes, gt_labels,
        iou_thresh=0.5):
    """Calculate precision and recall based on evaluation code of PASCAL VOC.
    This function calculates precision and recall of
    predicted bounding boxes obtained from a dataset which has :math:`N`
    images.
    The code is based on the evaluation code used in PASCAL VOC Challenge.
    Args:
        pred_bboxes (iterable of numpy.ndarray): An iterable of :math:`N`
            sets of bounding boxes.
            Its index corresponds to an index for the base dataset.
            Each element of :obj:`pred_bboxes` is a set of coordinates
            of bounding boxes. This is an array whose shape is :math:`(R, 4)`,
            where :math:`R` corresponds
            to the number of bounding boxes, which may vary among boxes.
            The second axis corresponds to
            :math:`y_{min}, x_{min}, y_{max}, x_{max}` of a bounding box.
        pred_labels (iterable of numpy.ndarray): An iterable of labels.
            Similar to :obj:`pred_bboxes`, its index corresponds to an
            index for the base dataset. Its length is :math:`N`.
        pred_scores (iterable of numpy.ndarray): An iterable of confidence
            scores for predicted bounding boxes. Similar to :obj:`pred_bboxes`,
            its index corresponds to an index for the base dataset.
            Its length is :math:`N`.
        gt_bboxes (iterable of numpy.ndarray): An iterable of ground truth
            bounding boxes
            whose length is :math:`N`. An element of :obj:`gt_bboxes` is a
            bounding box whose shape is :math:`(R, 4)`. Note that the number of
            bounding boxes in each image does not need to be same as the number
            of corresponding predicted boxes.
        gt_labels (iterable of numpy.ndarray): An iterable of ground truth
            labels which are organized similarly to :obj:`gt_bboxes`.
        gt_difficults (iterable of numpy.ndarray): An iterable of boolean
            arrays which is organized similarly to :obj:`gt_bboxes`.
            This tells whether the
            corresponding ground truth bounding box is difficult or not.
            By default, this is :obj:`None`. In that case, this function
            considers all bounding boxes to be not difficult.
        iou_thresh (float): A prediction is correct if its Intersection over
            Union with the ground truth is above this value..
    Returns:
        tuple of two lists:
        This function returns two lists: :obj:`prec` and :obj:`rec`.
        * :obj:`prec`: A list of arrays. :obj:`prec[l]` is precision \
            for class :math:`l`. If class :math:`l` does not exist in \
            either :obj:`pred_labels` or :obj:`gt_labels`, :obj:`prec[l]` is \
            set to :obj:`None`.
        * :obj:`rec`: A list of arrays. :obj:`rec[l]` is recall \
            for class :math:`l`. If class :math:`l` that is not marked as \
            difficult does not exist in \
            :obj:`gt_labels`, :obj:`rec[l]` is \
            set to :obj:`None`.
    """

    pred_bboxes = iter(pred_bboxes)
    pred_labels = iter(pred_labels)
    pred_scores = iter(pred_scores)
    gt_bboxes = iter(gt_bboxes)
    gt_labels = iter(gt_labels)

    n_pos = defaultdict(int)
    score = defaultdict(list)
    match = defaultdict(list)

    for pred_bbox, pred_label, pred_score, gt_bbox, gt_label in \
            six.moves.zip(
                pred_bboxes, pred_labels, pred_scores,
                gt_bboxes, gt_labels):


        for l in np.unique(np.concatenate((pred_label, gt_label)).astype(int)):
            pred_mask_l = pred_label == l
            pred_bbox_l = pred_bbox[pred_mask_l]
            pred_score_l = pred_score[pred_mask_l]
            # sort by score
            order = pred_score_l.argsort()[::-1]
            pred_bbox_l = pred_bbox_l[order]
            pred_score_l = pred_score_l[order]

            gt_mask_l = gt_label == l
            gt_bbox_l = gt_bbox[gt_mask_l]
            
            n_pos[l] += gt_bbox_l.shape[0]
            score[l].extend(pred_score_l)

            if len(pred_bbox_l) == 0:
                continue
            if len(gt_bbox_l) == 0:
                match[l].extend((0,) * pred_bbox_l.shape[0])
                continue

            # VOC evaluation follows integer typed bounding boxes.
            pred_bbox_l = pred_bbox_l.copy()
            pred_bbox_l[:, 2:] += 1
            gt_bbox_l = gt_bbox_l.copy()
            gt_bbox_l[:, 2:] += 1

            iou = bbox_iou(pred_bbox_l, gt_bbox_l)
            gt_index = iou.argmax(axis=1)
            # set -1 if there is no matching ground truth
            gt_index[iou.max(axis=1) < iou_thresh] = -1
            del iou

            selec = np.zeros(gt_bbox_l.shape[0], dtype=bool)
            
            for gt_idx in gt_index:
                if gt_idx >= 0:
                    if not selec[gt_idx]:
                        match[l].append(1)
                    else:
                        match[l].append(0)
                    selec[gt_idx] = True
                else:
                    match[l].append(0)
            
    for iter_ in (
            pred_bboxes, pred_labels, pred_scores,
            gt_bboxes, gt_labels):
        if next(iter_, None) is not None:
            raise ValueError('Length of input iterables need to be same.')

    n_fg_class = max(n_pos.keys()) + 1
    prec = [None] * n_fg_class
    rec = [None] * n_fg_class

    for l in n_pos.keys():
        score_l = np.array(score[l])
        match_l = np.array(match[l], dtype=np.int8)

        order = score_l.argsort()[::-1]
        match_l = match_l[order]

        tp = np.cumsum(match_l == 1)
        fp = np.cumsum(match_l == 0)

        # If an element of fp + tp is 0,
        # the corresponding element of prec[l] is nan.
        prec[l] = tp / (fp + tp)
        # If n_pos[l] is 0, rec[l] is None.
        if n_pos[l] > 0:
            rec[l] = tp / n_pos[l]

    return prec, rec


def calc_detection_voc_ap(prec, rec, use_07_metric=False):
    """Calculate average precisions based on evaluation code of PASCAL VOC.
    This function calculates average precisions
    from given precisions and recalls.
    The code is based on the evaluation code used in PASCAL VOC Challenge.
    Args:
        prec (list of numpy.array): A list of arrays.
            :obj:`prec[l]` indicates precision for class :math:`l`.
            If :obj:`prec[l]` is :obj:`None`, this function returns
            :obj:`numpy.nan` for class :math:`l`.
        rec (list of numpy.array): A list of arrays.
            :obj:`rec[l]` indicates recall for class :math:`l`.
            If :obj:`rec[l]` is :obj:`None`, this function returns
            :obj:`numpy.nan` for class :math:`l`.
        use_07_metric (bool): Whether to use PASCAL VOC 2007 evaluation metric
            for calculating average precision. The default value is
            :obj:`False`.
    Returns:
        ~numpy.ndarray:
        This function returns an array of average precisions.
        The :math:`l`-th value corresponds to the average precision
        for class :math:`l`. If :obj:`prec[l]` or :obj:`rec[l]` is
        :obj:`None`, the corresponding value is set to :obj:`numpy.nan`.
    """

    n_fg_class = len(prec)
    ap = np.empty(n_fg_class)
    for l in six.moves.range(n_fg_class):
        if prec[l] is None or rec[l] is None:
            ap[l] = np.nan
            continue

        if use_07_metric:
            # 11 point metric
            ap[l] = 0
            for t in np.arange(0., 1.1, 0.1):
                if np.sum(rec[l] >= t) == 0:
                    p = 0
                else:
                    p = np.max(np.nan_to_num(prec[l])[rec[l] >= t])
                ap[l] += p / 11
        else:
            # correct AP calculation
            # first append sentinel values at the end
            mpre = np.concatenate(([0], np.nan_to_num(prec[l]), [0]))
            mrec = np.concatenate(([0], rec[l], [1]))

            mpre = np.maximum.accumulate(mpre[::-1])[::-1]

            # to calculate area under PR curve, look for points
            # where X axis (recall) changes value
            i = np.where(mrec[1:] != mrec[:-1])[0]

            # and sum (\Delta recall) * prec
            ap[l] = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
    return ap


import torch
from PIL import Image, ImageDraw, ImageFont 
import colorsys
import torch
import numpy as np 
import os
from remo.utils import yxyx_yxhw
import matplotlib.pyplot as plt
from matplotlib import patches

__all__ = ["Img_with_labels", "Img_with_grid", "Img_with_obj_ctr", "Img_obj_ctr_grid", "Img_obj_ctr_grid_boxes"]

def Img_with_labels(img, bboxes, labels, labels_mapping, save_loc=None, path="./"):
    """ Outputs an image with bounding boxes and labels 
    
    img: A PIL image. 
    bboxes: numpy array: [N, 4] [y1, x1, y2, x2]
    lables: numpy array: [N]
    labels_mapping: dict
    
    Returns a PIL image
    """
    image = img.copy()
    hsv_tuples = [(x/(len(labels_mapping)+5), 1., 1.) for x in range(len(labels_mapping)+5)]
    colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
    colors = list(map(lambda x: (int(x[0]*255), int(x[1]*255), int(x[2]*255)), colors))
    
    font = ImageFont.truetype(font=path+"/FiraMono-Medium.otf",
                             size=np.floor(1e-1 * (image.size[1]+0.5)/2).astype("int32"))
    thickness = (image.size[0]+image.size[1]) // 600
    
    for i, c in enumerate(labels):
        box = bboxes[i]
        label = '{}'.format(labels_mapping[c])
        draw = ImageDraw.Draw(image)
        label_size = draw.textsize(label, font)
        
        ymin, xmin, ymax, xmax = box
        bottom = ymax
        top = ymin
        left = xmin
        right = xmax

        if top - label_size[1] >=0:
            text_origin = np.array([left, top-label_size[1]])
        else:
            text_origin = np.array([left, top+1])

        for i in range(thickness):
            draw.rectangle(
            [left+i, top+i, right-i, bottom-i], outline=colors[c])

        draw.rectangle([tuple(text_origin), tuple(text_origin+label_size)], fill=colors[c])
        draw.text(text_origin, label, fill=(0, 0, 0), font=font)
        del draw
    
    if save_loc is not None:
        image.save(save_loc, quality=90)
    else:
        return image
    
def Img_with_grid(img, grid=(32, 32)):
    """ Outputs a PIL image with grid
    
    img: PIL image 
    grid: tuple. The size of the grid. (h, w)
    """
    image = img.copy()
    
    h = image.size[1]//grid[0]
    w = image.size[0]//grid[1]
    
    hlines = np.array([[0, h+n*h, image.size[1], h+n*h] for n in range(image.size[1]//h)])
    vlines = np.array([[w+n*w, 0, w+n*w, image.size[0]] for n in range(image.size[0]//w)])
    
    for i in hlines:
        draw = ImageDraw.Draw(image)
        draw.line([i[0], i[1], i[2], i[3]], fill="red", width=2)
        del draw 
    
    for i in vlines:
        draw = ImageDraw.Draw(image)
        draw.line([i[0], i[1], i[2], i[3]], fill="red", width=2)
        del draw 
    
    return image

def Img_with_obj_ctr(img, bboxes, labels, labels_mapping):
    """Outputs Image with object centers
    
    img: PIL image 
    bboxes: numpy array [N, 4] [x_ctr, y_ctr, h, w]
    labels: numpy array [N, 1] 
    labels_mapping: A dict with keys as numbers and values as labels
    """
    
    image = img.copy()
    hsv_tuples = [(x/(len(labels_mapping)+5), 1., 1.) for x in range(len(labels_mapping)+5)] #+5 is hardcoded
    colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
    colors = list(map(lambda x: (int(x[0]*255), int(x[1]*255), int(x[2]*255)), colors))
    r = (image.size[0]+image.size[1]) // 600
    
    for c, i in enumerate(bboxes):
        label = labels[c]
        draw = ImageDraw.Draw(image)
        x, y = i[1], i[0]
        draw.ellipse([x-r, y-r, x+r, y+r], fill=colors[label])
        del draw
        
    return image

def Img_obj_ctr_grid(img, bboxes, labels, labels_mapping, grid=(32, 32)):
    """ Combining Img_with_obj_ctr and Img_with_grid
    
    img: PIL image 
    bboxes: numpy array [N, 4]
    labels: numpy array [N, 1] 
    labels_mapping: A dict with keys as numbers and values as labels
    grid: tuple. The size of the grid. (h, w)
    """
    image = img.copy()
    image = Img_with_grid(image, grid)
    box = yxyx_yxhw(bboxes)
    image = Img_with_obj_ctr(image, box, labels, labels_mapping)
    return image


def Img_obj_ctr_grid_boxes(img, bboxes, labels, labels_mapping, grid=(32, 32), path="./", save_loc=None):
    """ Combining Img_with_obj_ctr and Img_with_grid
    
    img: PIL image 
    bboxes: numpy array [N, 4]
    labels: numpy array [N, 1] 
    labels_mapping: A dict with keys as numbers and values as labels
    grid: tuple. The size of the grid. (h, w)
    """
    image = img.copy()
    image = Img_with_grid(image, grid)
    box = yxyx_yxhw(bboxes)
    image = Img_with_obj_ctr(image, box, labels, labels_mapping)
    image = Img_with_labels(image, bboxes, labels, labels_mapping, save_loc, path)
    return image


def draw_bounding_boxes(image, bbox, line_color="blue", linewidth=2, figsize=(30, 20), fill=None):
    fig = plt.figure(figsize=figsize)
    ax = fig.subplots(1)
    ax.imshow(image)
    for i in bbox:
        top = i[0]
        left = i[1]
        height = i[2]
        width = i[3]
        rect = patches.Rectangle([left, top], width, height, fill=fill)
        ax.add_patch(rect)
    plt.show()
    return ax 

def draw_grid(image, bbox, outline="white"):
    draw = ImageDraw.Draw(image)
    for i in bbox:
        y0, x0, y1, x1 = i
        draw.rectangle([x0, y0, x1, y1], outline=outline)
    return image


__author__ = ["Not me"]

import torch
import numpy as np 
import torch.nn as nn


def bbox_iou(box1, box2):
    """ Returns the IoU of two bounding boxes 
    """
    #Get the coordinates of bounding boxes
    b1_x1, b1_y1, b1_x2, b1_y2 = box1[:,0], box1[:,1], box1[:,2], box1[:,3]
    b2_x1, b2_y1, b2_x2, b2_y2 = box2[:,0], box2[:,1], box2[:,2], box2[:,3]
    
    #get the corrdinates of the intersection rectangle
    inter_rect_x1 =  torch.max(b1_x1, b2_x1)
    inter_rect_y1 =  torch.max(b1_y1, b2_y1)
    inter_rect_x2 =  torch.min(b1_x2, b2_x2)
    inter_rect_y2 =  torch.min(b1_y2, b2_y2)
    
    #Intersection area
    inter_area = torch.clamp(inter_rect_x2 - inter_rect_x1 + 1, min=0) * torch.clamp(inter_rect_y2 - inter_rect_y1 + 1, min=0)
    
    #Union Area
    b1_area = (b1_x2 - b1_x1 + 1)*(b1_y2 - b1_y1 + 1)
    b2_area = (b2_x2 - b2_x1 + 1)*(b2_y2 - b2_y1 + 1)
    
    iou = inter_area / (b1_area + b2_area - inter_area)
    
    return iou

def unique(tensor):
    tensor_np = tensor.cpu().numpy()
    unique_np = np.unique(tensor_np)
    unique_tensor = torch.from_numpy(unique_np)
    
    tensor_res = tensor.new(unique_tensor.shape)
    tensor_res.copy_(unique_tensor)
    return tensor_res


def write_results(prediction, confidence, num_classes, nms = True, nms_conf = 0.4):
    conf_mask = (prediction[:,:,4] > confidence).float().unsqueeze(2)
    prediction = prediction*conf_mask
    

    try:
        ind_nz = torch.nonzero(prediction[:,:,4]).transpose(0,1).contiguous()
    except:
        return 0
    
    
    box_a = prediction.new(prediction.shape)
    box_a[:,:,0] = (prediction[:,:,0] - prediction[:,:,2]/2)
    box_a[:,:,1] = (prediction[:,:,1] - prediction[:,:,3]/2)
    box_a[:,:,2] = (prediction[:,:,0] + prediction[:,:,2]/2) 
    box_a[:,:,3] = (prediction[:,:,1] + prediction[:,:,3]/2)
    prediction[:,:,:4] = box_a[:,:,:4]
    

    
    batch_size = prediction.size(0)
    
    output = prediction.new(1, prediction.size(2) + 1)
    write = False


    for ind in range(batch_size):
        #select the image from the batch
        image_pred = prediction[ind]
        

        
        #Get the class having maximum score, and the index of that class
        #Get rid of num_classes softmax scores 
        #Add the class index and the class score of class having maximum score
        max_conf, max_conf_score = torch.max(image_pred[:,5:5+ num_classes], 1)
        max_conf = max_conf.float().unsqueeze(1)
        max_conf_score = max_conf_score.float().unsqueeze(1)
        seq = (image_pred[:,:5], max_conf, max_conf_score)
        image_pred = torch.cat(seq, 1)
        

        
        #Get rid of the zero entries
        non_zero_ind =  (torch.nonzero(image_pred[:,4]))

        
        image_pred_ = image_pred[non_zero_ind.squeeze(),:].view(-1,7)
        
        #Get the various classes detected in the image
        try:
            img_classes = unique(image_pred_[:,-1])
        except:
             continue
        #WE will do NMS classwise
        for cls in img_classes:
            #get the detections with one particular class
            cls_mask = image_pred_*(image_pred_[:,-1] == cls).float().unsqueeze(1)
            class_mask_ind = torch.nonzero(cls_mask[:,-2]).squeeze()
            

            image_pred_class = image_pred_[class_mask_ind].view(-1,7)

		
        
             #sort the detections such that the entry with the maximum objectness
             #confidence is at the top
            conf_sort_index = torch.sort(image_pred_class[:,4], descending = True )[1]
            image_pred_class = image_pred_class[conf_sort_index]
            idx = image_pred_class.size(0)
            
            #if nms has to be done
            if nms:
                #For each detection
                for i in range(idx):
                    #Get the IOUs of all boxes that come after the one we are looking at 
                    #in the loop
                    try:
                        ious = bbox_iou(image_pred_class[i].unsqueeze(0), image_pred_class[i+1:])
                    except ValueError:
                        break
        
                    except IndexError:
                        break
                    
                    #Zero out all the detections that have IoU > treshhold
                    iou_mask = (ious < nms_conf).float().unsqueeze(1)
                    image_pred_class[i+1:] *= iou_mask       
                    
                    #Remove the non-zero entries
                    non_zero_ind = torch.nonzero(image_pred_class[:,4]).squeeze()
                    image_pred_class = image_pred_class[non_zero_ind].view(-1,7)
                    
                    

            #Concatenate the batch_id of the image to the detection
            #this helps us identify which image does the detection correspond to 
            #We use a linear straucture to hold ALL the detections from the batch
            #the batch_dim is flattened
            #batch is identified by extra batch column
            
            
            batch_ind = image_pred_class.new(image_pred_class.size(0), 1).fill_(ind)
            seq = batch_ind, image_pred_class
            if not write:
                output = torch.cat(seq,1)
                write = True
            else:
                out = torch.cat(seq,1)
                output = torch.cat((output,out))
    
    return output


def xyxy_xywh(boxes, offset=0.5):
    """xyxy to xywh format  
    """
    height = boxes[:, 3] - boxes[:, 1]
    width = boxes[:, 2] - boxes[:, 0]
    ctr_y = boxes[:, 1] + offset * height
    ctr_x = boxes[:, 0] + offset * width 
    return torch.stack((ctr_x, ctr_y, width, height), dim=1)
        
#def xyxy_xyhw(boxes, offset=0.5):
#    """xyxy
#    """
#    height = boxes[:, 3] - boxes[:, 1]
#    width = boxes[:, 2] - boxes[:, 0]
#    ctr_y = boxes[:, 1] + offset * height
#    ctr_x = boxes[:, 0] + offset * width 
#    return torch.stack((ctr_x, ctr_y, height, width), dim=1)

def clamper(bbox, size):
    """bbox takes xyxy values
    """
    h, w = size
    bbox[:, 0] = torch.clamp(bbox[:, 0], 0, w)
    bbox[:, 1] = torch.clamp(bbox[:, 1], 0, h)
    bbox[:, 2] = torch.clamp(bbox[:, 2], 0, w)
    bbox[:, 3] = torch.clamp(bbox[:, 3], 0, h)
    return bbox

def xywh_xyxy(boxes):
    """xyhw_xyxy ----> we are using h,w and not w, h format .. Please be careful
    """
    bbox = torch.zeros(boxes.shape)
    bbox[:, 0] = boxes[:, 0] - 0.5 * boxes[:, 2]
    bbox[:, 1] = boxes[:, 1] - 0.5 * boxes[:, 3]
    bbox[:, 2] = boxes[:, 0] + 0.5 * boxes[:, 2]
    bbox[:, 3] = boxes[:, 1] + 0.5 * boxes[:, 3]
    return bbox

    
#def xyhw_xyxy(boxes):
#    """xyhw_xyxy
#    """
#    bbox = torch.zeros(boxes.shape)
#    bbox[:, 0] = boxes[:, 0] - 0.5 * boxes[:, 3]
#    bbox[:, 1] = boxes[:, 1] - 0.5 * boxes[:, 2]
#    bbox[:, 2] = boxes[:, 0] + 0.5 * boxes[:, 3]
#    bbox[:, 3] = boxes[:, 1] + 0.5 * boxes[:, 2]
#    return bbox

def reframe_bboxes_for_actual_image(img_size, resized_img_size, bboxes):
    h, w = img_size
    oh, ow = resized_img_size

    sw = w/ float(ow) 
    sh = h/ float(oh)
    
    xywh = xyxy_xywh(bboxes)
    xywh[:, 2] = (w/float(ow)) * xywh[:, 2]
    xywh[:, 3] = (h/float(oh)) * xywh[:, 3]
    xywh[:, 1] = (h/float(oh)) * xywh[:, 1]
    xywh[:, 0] = (w/float(ow)) * xywh[:, 0]
    
    bbox = xywh_xyxy(xywh)
    return bbox


# vim: expandtab:ts=4:sw=4
from __future__ import absolute_import
import numpy as np
from sklearn.utils.linear_assignment_ import linear_assignment
from . import kalman_filter


INFTY_COST = 1e+5


def min_cost_matching(
        distance_metric, max_distance, tracks, detections, track_indices=None,
        detection_indices=None):
    """Solve linear assignment problem.

    Parameters
    ----------
    distance_metric : Callable[List[Track], List[Detection], List[int], List[int]) -> ndarray
        The distance metric is given a list of tracks and detections as well as
        a list of N track indices and M detection indices. The metric should
        return the NxM dimensional cost matrix, where element (i, j) is the
        association cost between the i-th track in the given track indices and
        the j-th detection in the given detection_indices.
    max_distance : float
        Gating threshold. Associations with cost larger than this value are
        disregarded.
    tracks : List[track.Track]
        A list of predicted tracks at the current time step.
    detections : List[detection.Detection]
        A list of detections at the current time step.
    track_indices : List[int]
        List of track indices that maps rows in `cost_matrix` to tracks in
        `tracks` (see description above).
    detection_indices : List[int]
        List of detection indices that maps columns in `cost_matrix` to
        detections in `detections` (see description above).

    Returns
    -------
    (List[(int, int)], List[int], List[int])
        Returns a tuple with the following three entries:
        * A list of matched track and detection indices.
        * A list of unmatched track indices.
        * A list of unmatched detection indices.

    """
    if track_indices is None:
        track_indices = np.arange(len(tracks))
    if detection_indices is None:
        detection_indices = np.arange(len(detections))

    if len(detection_indices) == 0 or len(track_indices) == 0:
        return [], track_indices, detection_indices  # Nothing to match.

    cost_matrix = distance_metric(
        tracks, detections, track_indices, detection_indices)
    cost_matrix[cost_matrix > max_distance] = max_distance + 1e-5
    indices = linear_assignment(cost_matrix)

    matches, unmatched_tracks, unmatched_detections = [], [], []
    for col, detection_idx in enumerate(detection_indices):
        if col not in indices[:, 1]:
            unmatched_detections.append(detection_idx)
    for row, track_idx in enumerate(track_indices):
        if row not in indices[:, 0]:
            unmatched_tracks.append(track_idx)
    for row, col in indices:
        track_idx = track_indices[row]
        detection_idx = detection_indices[col]
        if cost_matrix[row, col] > max_distance:
            unmatched_tracks.append(track_idx)
            unmatched_detections.append(detection_idx)
        else:
            matches.append((track_idx, detection_idx))
    return matches, unmatched_tracks, unmatched_detections


def matching_cascade(
        distance_metric, max_distance, cascade_depth, tracks, detections,
        track_indices=None, detection_indices=None):
    """Run matching cascade.

    Parameters
    ----------
    distance_metric : Callable[List[Track], List[Detection], List[int], List[int]) -> ndarray
        The distance metric is given a list of tracks and detections as well as
        a list of N track indices and M detection indices. The metric should
        return the NxM dimensional cost matrix, where element (i, j) is the
        association cost between the i-th track in the given track indices and
        the j-th detection in the given detection indices.
    max_distance : float
        Gating threshold. Associations with cost larger than this value are
        disregarded.
    cascade_depth: int
        The cascade depth, should be se to the maximum track age.
    tracks : List[track.Track]
        A list of predicted tracks at the current time step.
    detections : List[detection.Detection]
        A list of detections at the current time step.
    track_indices : Optional[List[int]]
        List of track indices that maps rows in `cost_matrix` to tracks in
        `tracks` (see description above). Defaults to all tracks.
    detection_indices : Optional[List[int]]
        List of detection indices that maps columns in `cost_matrix` to
        detections in `detections` (see description above). Defaults to all
        detections.

    Returns
    -------
    (List[(int, int)], List[int], List[int])
        Returns a tuple with the following three entries:
        * A list of matched track and detection indices.
        * A list of unmatched track indices.
        * A list of unmatched detection indices.

    """
    if track_indices is None:
        track_indices = list(range(len(tracks)))
    if detection_indices is None:
        detection_indices = list(range(len(detections)))

    unmatched_detections = detection_indices
    matches = []
    for level in range(cascade_depth):
        if len(unmatched_detections) == 0:  # No detections left
            break

        track_indices_l = [
            k for k in track_indices
            if tracks[k].time_since_update == 1 + level
        ]
        if len(track_indices_l) == 0:  # Nothing to match at this level
            continue

        matches_l, _, unmatched_detections = \
            min_cost_matching(
                distance_metric, max_distance, tracks, detections,
                track_indices_l, unmatched_detections)
        matches += matches_l
    unmatched_tracks = list(set(track_indices) - set(k for k, _ in matches))
    return matches, unmatched_tracks, unmatched_detections


def gate_cost_matrix(
        kf, cost_matrix, tracks, detections, track_indices, detection_indices,
        gated_cost=INFTY_COST, only_position=False):
    """Invalidate infeasible entries in cost matrix based on the state
    distributions obtained by Kalman filtering.

    Parameters
    ----------
    kf : The Kalman filter.
    cost_matrix : ndarray
        The NxM dimensional cost matrix, where N is the number of track indices
        and M is the number of detection indices, such that entry (i, j) is the
        association cost between `tracks[track_indices[i]]` and
        `detections[detection_indices[j]]`.
    tracks : List[track.Track]
        A list of predicted tracks at the current time step.
    detections : List[detection.Detection]
        A list of detections at the current time step.
    track_indices : List[int]
        List of track indices that maps rows in `cost_matrix` to tracks in
        `tracks` (see description above).
    detection_indices : List[int]
        List of detection indices that maps columns in `cost_matrix` to
        detections in `detections` (see description above).
    gated_cost : Optional[float]
        Entries in the cost matrix corresponding to infeasible associations are
        set this value. Defaults to a very large value.
    only_position : Optional[bool]
        If True, only the x, y position of the state distribution is considered
        during gating. Defaults to False.

    Returns
    -------
    ndarray
        Returns the modified cost matrix.

    """
    gating_dim = 2 if only_position else 4
    gating_threshold = kalman_filter.chi2inv95[gating_dim]
    measurements = np.asarray(
        [detections[i].to_xyah() for i in detection_indices])
    for row, track_idx in enumerate(track_indices):
        track = tracks[track_idx]
        gating_distance = kf.gating_distance(
            track.mean, track.covariance, measurements, only_position)
        cost_matrix[row, gating_distance > gating_threshold] = gated_cost
    return cost_matrix



import torchvision
import torch.nn as nn

__models__ = ["densnet121", "densenet161", "densenet169", "densenet201"]


def densnet121(num_classes=1000, pretrained='imagenet'):
    if pretrained == "imagenet":
        model = torchvision.model.densnet121(pretrained=True)
        if num_classes != 1000:
            model.classifier = nn.Linear(model.classifier.in_features, num_classes)
    else:
        model = torchvision.models.densnet121(pretrained=False, num_classes=num_classes)
    return model

def densnet161(num_classes=1000, pretrained='imagenet'):
    if pretrained == "imagenet":
        model = torchvision.model.densnet161(pretrained=True)
        if num_classes != 1000:
            model.classifier = nn.Linear(model.classifier.in_features, num_classes)
    else:
        model = torchvision.models.densnet161(pretrained=False, num_classes=num_classes)
    return model

def densnet169(num_classes=1000, pretrained='imagenet'):
    if pretrained == "imagenet":
        model = torchvision.model.densnet169(pretrained=True)
        if num_classes != 1000:
            model.classifier = nn.Linear(model.classifier.in_features, num_classes)
    else:
        model = torchvision.models.densnet169(pretrained=False, num_classes=num_classes)
    return model

def densnet201(num_classes=1000, pretrained='imagenet'):
    if pretrained == "imagenet":
        model = torchvision.model.densnet201(pretrained=True)
        if num_classes != 1000:
            model.classifier = nn.Linear(model.classifier.in_features, num_classes)
    else:
        model = torchvision.models.densnet201(pretrained=False, num_classes=num_classes)
    return model



import torch
import torch.nn as nn
import torch.nn.functional as F

class BasicBlock(nn.Module):
    def __init__(self, c_in, c_out,is_downsample=False):
        super(BasicBlock,self).__init__()
        self.is_downsample = is_downsample
        if is_downsample:
            self.conv1 = nn.Conv2d(c_in, c_out, 3, stride=2, padding=1, bias=False)
        else:
            self.conv1 = nn.Conv2d(c_in, c_out, 3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(c_out)
        self.relu = nn.ReLU(True)
        self.conv2 = nn.Conv2d(c_out,c_out,3,stride=1,padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(c_out)
        if is_downsample:
            self.downsample = nn.Sequential(
                nn.Conv2d(c_in, c_out, 1, stride=2, bias=False),
                nn.BatchNorm2d(c_out)
            )
        elif c_in != c_out:
            self.downsample = nn.Sequential(
                nn.Conv2d(c_in, c_out, 1, stride=1, bias=False),
                nn.BatchNorm2d(c_out)
            )
            self.is_downsample = True

    def forward(self,x):
        y = self.conv1(x)
        y = self.bn1(y)
        y = self.relu(y)
        y = self.conv2(y)
        y = self.bn2(y)
        if self.is_downsample:
            x = self.downsample(x)
        return F.relu(x.add(y),True)

def make_layers(c_in,c_out,repeat_times, is_downsample=False):
    blocks = []
    for i in range(repeat_times):
        if i ==0:
            blocks += [BasicBlock(c_in,c_out, is_downsample=is_downsample),]
        else:
            blocks += [BasicBlock(c_out,c_out),]
    return nn.Sequential(*blocks)

class Net(nn.Module):
    def __init__(self, num_classes=751 ,reid=False):
        super(Net,self).__init__()
        # 3 128 64
        self.conv = nn.Sequential(
            nn.Conv2d(3,64,3,stride=1,padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
            # nn.Conv2d(32,32,3,stride=1,padding=1),
            # nn.BatchNorm2d(32),
            # nn.ReLU(inplace=True),
            nn.MaxPool2d(3,2,padding=1),
        )
        # 32 64 32
        self.layer1 = make_layers(64,64,2,False)
        # 32 64 32
        self.layer2 = make_layers(64,128,2,True)
        # 64 32 16
        self.layer3 = make_layers(128,256,2,True)
        # 128 16 8
        self.layer4 = make_layers(256,512,2,True)
        # 256 8 4
        self.avgpool = nn.AvgPool2d((8,4),1)
        # 256 1 1 
        self.reid = reid
        self.classifier = nn.Sequential(
            nn.Linear(512, 256),
            nn.BatchNorm1d(256),
            nn.ReLU(inplace=True),
            nn.Dropout(),
            nn.Linear(256, num_classes),
        )
    
    def forward(self, x):
        x = self.conv(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.avgpool(x)
        x = x.view(x.size(0),-1)
        # B x 128
        if self.reid:
            x = x.div(x.norm(p=2,dim=1,keepdim=True))
            return x
        # classifier
        x = self.classifier(x)
        return x


if __name__ == '__main__':
    net = Net()
    x = torch.randn(4,3,128,64)
    y = net(x)
    import ipdb; ipdb.set_trace()





import torch

def xyxy_xywh(boxes, offset=0.5):
    """xyxy
    """
    height = boxes[:, 3] - boxes[:, 1]
    width = boxes[:, 2] - boxes[:, 0]
    ctr_y = boxes[:, 1] + offset * height
    ctr_x = boxes[:, 0] + offset * width 
    return torch.stack((ctr_x, ctr_y, width, height), dim=1)


import torch
import torchvision.transforms as transforms
import numpy as np
import cv2

from .model import Net

class Extractor(object):
    def __init__(self, model_path, use_cuda=True):
        self.net = Net(reid=True)
        self.device = "cuda" if torch.cuda.is_available() and use_cuda else "cpu"
        state_dict = torch.load(model_path)['net_dict']
        self.net.load_state_dict(state_dict)
        print("Loading weights from {}... Done!".format(model_path))
        self.net.to(self.device)
        self.norm = transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])

    def __call__(self, img):
        assert isinstance(img, np.ndarray), "type error"
        img = img.astype(np.float)#/255.
        print(img.size)
        print(img)
        img = cv2.resize(img, (64,128))
        img = torch.from_numpy(img).float().permute(2,0,1)
        img = self.norm(img).unsqueeze(0)
        with torch.no_grad():
            img = img.to(self.device)
            feature = self.net(img)
        return feature.cpu().numpy()


if __name__ == '__main__':
    img = cv2.imread("../../demo.jpg")[:,:,(2,1,0)]
    extr = Extractor("../../../../deep_sort_pytorch/deep/checkpoint/ckpt.t7")
    feature = extr(img)
    print(feature.shape)




""" For now write individual anchor boxes generator
"""
import torch
import numpy as np 

from person_tracking.fractal.anchors.utils import bbox_iou, xywh_xyxy, clamper, xyxy_xywh
# from fractal.anchors.utils import bbox_iou, xywh_xyxy, clamper, xyxy_xywh


def get_inside_index(bbox, H, W):
    """ index of the elements whose co-ordinates are bound inside the image
    """
    index = torch.where((bbox[:, 0] >= 0) &
        (bbox[:, 1] >= 0) &
        (bbox[:, 2] <= H) &
        (bbox[:, 3] <= W), torch.ones(bbox.shape[0]), torch.zeros(bbox.shape[0])).nonzero()
    return index.squeeze(1)

def _unmap(data, count, index, fill=-1):
    if len(data.shape) == 1:
        t = torch.zeros(count)
        t.fill_(fill)
        t[index] = data
    else:
        t = torch.zeros((count, data.shape[1:][0]))
        t.fill_(fill)
        t[index, :] = data
    return t 


def generate_grid(grid_size, offset=0.5):
    """
    :param grid_size:
    :param offset: offset from left top corner
    :return:
    """
    grid = np.arange(grid_size)
    x_offset, y_offset = np.meshgrid(grid, grid)
    x_offset, y_offset = torch.FloatTensor(x_offset) + offset, torch.FloatTensor(y_offset) + offset
    grid_cells = torch.cat((x_offset.unsqueeze(2), y_offset.unsqueeze(2)), 2)
    grid_cells = grid_cells.view(-1, 2)
    return grid_cells


def generate_anchors(stride, nW, nH, anchors_scales):
    grid_size_x = nW/stride
    grid_size_y = nH/stride
    grid = generate_grid(grid_size_x, 0)
    wh_anchors = anchors_scales /stride
    all_grid_xy = grid.repeat(1, 3).view(-1, 2)
    all_grid_wh = wh_anchors.repeat(int(grid_size_x*grid_size_y), 1)
    return torch.cat((all_grid_xy, all_grid_wh), 1)


class Yolov3:
    
    def __init__(self, strides, anchor_centroids, img_size):
        """

        :param strides: [32, 16, 8]
        :param anchor_centroids:  [[[10, 13], [16, 30], [33, 23]],
                                   [[30, 61], [62, 45], [59, 119]],
                                   [[116, 90], [156, 198], [373, 326]]]
                                   Anchor box attr per centroid (here shape is 3X3X2)
        :param img_size: 416
        """
        self.strides = strides
        self.anchor_centroids = anchor_centroids
        self.img_size = img_size
        self.anchors_attr_list, self.anchor_strides_list = self.generate_anchors() # cx,cy,w,h format
        self.all_anchors = torch.cat(self.anchors_attr_list)
        self.all_anchors_xy = xywh_xyxy(self.all_anchors)
        self.all_anchors_strides = torch.cat(self.anchor_strides_list)
        self.all_anchors_ft_scale = self.get_anchors_ft_scale()
        self.valid_anchors_index = self.valid_index()
        self.anchor_tiles = self.anchor_tiles()
        self.valid_anchor_tiles = self.anchor_tiles[self.valid_anchors_index]
    
    def anchor_tiles(self):
        all_anchors_ft_scale = self.all_anchors_ft_scale
        all_anchors_ft_scale[:, 0] = torch.floor(all_anchors_ft_scale[:, 0])
        all_anchors_ft_scale[:, 1] = torch.floor(all_anchors_ft_scale[:, 1])
        return torch.cat([all_anchors_ft_scale[:, :2], 1+all_anchors_ft_scale[:, :2]], 1)
    
    def valid_index(self):
        all_anchors_xyxy = xywh_xyxy(self.all_anchors)
        valid_index = get_inside_index(all_anchors_xyxy, self.img_size, self.img_size)
        return valid_index
        
    def get_anchors_ft_scale(self):
        return self.all_anchors / self.all_anchors_strides.view(-1, 1).repeat(1, 4)
    
    @staticmethod
    def generate_grid(grid_size, offset=0.5):
        """
        :param grid_size:
        :param offset: offset from left top corner
        :return:
        """
        grid = np.arange(grid_size)
        x_offset, y_offset = np.meshgrid(grid, grid)
        x_offset, y_offset = torch.FloatTensor(x_offset) + offset, torch.FloatTensor(y_offset) + offset
        grid_cells = torch.cat((x_offset.unsqueeze(2), y_offset.unsqueeze(2)), 2)
        grid_cells = grid_cells.view(-1, 2)
        return grid_cells


    def generate_anchors(self):
        anchor_attrs_list = []
        anchor_strides = []
        for num, stride in enumerate(self.strides):
            grid_size = self.img_size // stride #13
            grid_cells = Yolov3.generate_grid(grid_size) # 13*13 X 2
            anchors_for_stride = self.anchor_centroids[num] #[[10, 13], [16, 30], [33, 23]]
            grid_cells_r = grid_cells.repeat(1, len(anchors_for_stride)).view(-1, 2) # 13*13*3 X2                             
            grid_cells_r *= stride # cordinates on  image_scale                                 
            anchors_r = torch.Tensor(anchors_for_stride).repeat(grid_size*grid_size, 1) # 3*13*13X2
            anchors_attrs_iscale = torch.cat((grid_cells_r, anchors_r),1) #13*13*3 X 4
            anchor_attrs_list.append(anchors_attrs_iscale)
            anchor_strides.append((torch.ones(anchors_attrs_iscale.shape[0])).fill_(stride))
            
        return anchor_attrs_list, anchor_strides
    
    def generate_pos_anchor_boxes_mask(self, bbox):
        """bbox will be in xywh format
        """
        total_gt = bbox.shape[0]
        strides = self.all_anchors_strides.view(-1, 1).repeat(total_gt, 1, 4)
        bbox_re = bbox.unsqueeze(1).repeat(1, 1, self.all_anchors.shape[0]).view(total_gt, -1, 4)
        bbox_re_strides = bbox_re/strides
        
        where = ((bbox_re_strides[:, :, 0] > self.anchor_tiles[:, 0]) &
                 (bbox_re_strides[:, :, 0] < self.anchor_tiles[:, 2]) &
                 (bbox_re_strides[:, :, 1] > self.anchor_tiles[:, 1]) &
                 (bbox_re_strides[:, :, 1] < self.anchor_tiles[:, 3]))
        where = where.permute(1, 0)
        where = where[self.valid_anchors_index]
        return where
    
    def encoder(self, bbox, labels):
        """ bbox is in xywh mode
        """
        ##Objectness score
        bbox_xywh = bbox 
        bbox = xywh_xyxy(bbox_xywh)
        assigned_anchors = self.generate_pos_anchor_boxes_mask(bbox_xywh) #[valid_anchors, gt_boxes]
        ious = bbox_iou(self.all_anchors_xy[self.valid_anchors_index][:, [1, 0, 3, 2]].numpy(), bbox[:, [1, 0, 3, 2]].numpy())
        
        single_pos = np.multiply(ious, assigned_anchors.float().numpy()).argmax(0)
        ignore_pos= np.where(ious > 0.5)[0]
        
        #if len(ignore_pos) == 0:
        #    ious_max = 0
        #    single_pos =0 
            
        #else:
        #    single_pos = ignore_pos[assigned_anchors[ignore_pos].argmax(0).numpy()]
        #    which_bbox = which_bbox[assigned_anchors[ignore_pos].argmax(0).numpy()] ## Will give in that order
        
        os_score = torch.zeros(self.valid_anchors_index.shape[0])
        os_score.fill_(0)
        if len(ignore_pos) > 0:
            os_score[ignore_pos] = -1 
        os_score[single_pos] = 1 
        
        f_os_score = _unmap(os_score, self.all_anchors.shape[0], self.valid_anchors_index, -1)
        
        ## xyhw 
        
        xy_anchors = self.all_anchors[self.valid_anchors_index][single_pos]
        strides = self.all_anchors_strides[self.valid_anchors_index][single_pos].view(-1, 1).repeat(1, 2)
        #xy_bboxes = bbox[which_bbox] #IF which_bbox is [2, 1, 0] incase of 3 boxes . this will change the order
        
        xy_anchors_tile = torch.floor(xy_anchors[:, :2]/strides)
        xy_bboxes_tile = bbox_xywh[:, :2]/strides
        
        xy_score = xy_bboxes_tile - xy_anchors_tile
        wh_score = torch.log(bbox_xywh[:, 2:]/xy_anchors[:, 2:])
        
        xywh_score = torch.cat([xy_score, wh_score], 1)
        
        coord_score = torch.zeros(self.valid_anchors_index.shape[0], 4)
        coord_score.fill_(-1) 
        coord_score[single_pos] = xywh_score 
        
        f_coord_score = _unmap(coord_score, self.all_anchors.shape[0], self.valid_anchors_index, -1)
        
        ## labels 
        l_score = torch.zeros(self.valid_anchors_index.shape[0])
        l_score.fill_(-1)
        l_score[single_pos] = labels.float()
        
        f_l_score = _unmap(l_score, self.all_anchors.shape[0], self.valid_anchors_index, -1)
        return torch.cat([f_coord_score, f_os_score.view(-1, 1), f_l_score.view(-1, 1)], 1)
        
            
            
            
        


import torch
from torchvision.transforms import transforms
import numpy as np
import visdom

import matplotlib.pyplot as plt 
from PIL import Image


class VisdomLinePlotter(object):
    """Plots to Visdom
    """
    def __init__(self, env_name='main'):
        self.viz = visdom.Visdom()
        self.env = env_name
        self.plots = {}
    def plot(self, var_name, split_name, x, y):
        """ Plot all the metrics available as named tuple
        
        var_name: named tuple 
        """
        if var_name not in self.plots:
            self.plots[var_name] = self.viz.line(X=np.array([x,x]), Y=np.array([y,y]), env=self.env, opts=dict(
                legend=[split_name],
                title=var_name,
                xlabel='Iterations',
                ylabel=var_name
            ))
        else:
            self.viz.line(X=np.array([x]), Y=np.array([y]), env=self.env, win=self.plots[var_name], name=split_name, update = 'append')

    def plot_tuple(self, ntuple, name, iteration):

        for field, value in ntuple._asdict().items():
            self.plot(field, name, iteration, value.data.cpu().numpy()[0])


def vis_images(images_tensor, labels, imgs_per_row=4, mode="RGB", fig_size= (20, 3)):
    if isinstance(images_tensor, list):
        pil_images = [Image.open(i) for i in images_tensor]
    else:
        pil_convertor = transforms.ToPILImage(mode=mode)
        pil_images=[pil_convertor(i) for i in images_tensor]
    if labels is None:
        labels= ["No label" for i in range(len(pil_images))]
    else:
        labels = [i[0] for i in labels.numpy()]
    
    batches = len(pil_images)//imgs_per_row
    for i in range(batches):
        imgs = pil_images[i*imgs_per_row:(i+1)*imgs_per_row]
        lab = labels[i*imgs_per_row:(i+1)*imgs_per_row]
        fig, ax = plt.subplots(nrows=1, ncols=4, sharex="col", sharey="row", figsize=fig_size)
        for i, img in enumerate(imgs):
            ax[i].imshow(img)
            ax[i].set_title(str(lab[i]))


# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['person_tracking',
 'person_tracking.fractal.anchors',
 'person_tracking.fractal.common_utils',
 'person_tracking.fractal.deep_sort',
 'person_tracking.fractal.loaders',
 'person_tracking.fractal.losses',
 'person_tracking.fractal.metrics',
 'person_tracking.fractal.nets',
 'person_tracking.fractal.optims',
 'person_tracking.fractal.superai',
 'person_tracking.fractal.superai.Nets',
 'person_tracking.fractal.superai.Nets.resnext_features',
 'person_tracking.fractal.vis_lib']

package_data = \
{'': ['*'],
 'person_tracking': ['cfgs/*', 'data/coco/*'],
 'person_tracking.fractal.superai': ['.ipynb_checkpoints/*']}

install_requires = \
['PILLOW>=7.0,<8.0',
 'matplotlib>=3.1,<4.0',
 'opencv-python>=4.1,<5.0',
 'progressbar>=2.5,<3.0',
 'scipy>=1.4,<2.0',
 'sklearn>=0.0.0,<0.0.1',
 'torch>=1.3,<2.0',
 'torchvision>=0.5.0,<0.6.0']

setup_kwargs = {
    'name': 'person-tracking',
    'version': '0.1.7',
    'description': 'person tracking',
    'long_description': None,
    'author': 'Sindhura',
    'author_email': 'sindhura.k@fractal.ai',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '==3.6.9',
}


setup(**setup_kwargs)



import torch
from torchvision import transforms
from person_tracking.fractal.loaders import ListDataset
from person_tracking.fractal.vis_lib.vis_utils import save_img_with_labels
from person_tracking.fractal.common_utils.img_utils import reframe_bboxes_for_actual_image, clamper
from person_tracking.fractal.nets.darknet import Darknet_yolo
from person_tracking.fractal.anchors import get_anchors
# from fractal.loaders import ListDataset
# from fractal.vis_lib.vis_utils import save_img_with_labels
# from fractal.common_utils.img_utils import reframe_bboxes_for_actual_image, clamper
# from fractal.nets.darknet import Darknet_yolo
# from fractal.anchors import get_anchors
import os
import warnings
import numpy as np
warnings.filterwarnings('ignore')
import time
from PIL import Image
import cv2
import progressbar
import sys
# import dlib
import matplotlib.pyplot as plt
# from heatmappy import Heatmapper
#pip install heatmappy
from person_tracking.fractal.deep_sort import preprocessing
from person_tracking.fractal.deep_sort import nn_matching
from person_tracking.fractal.deep_sort.detection import Detection
from person_tracking.fractal.deep_sort.tracker import Tracker
from person_tracking.fractal.deep_sort.detection import Detection as ddet
from person_tracking.fractal.deep_sort.feature_extractor import Extractor

from person_tracking.fractal.superai.Nets.resnets import resnet18

# from fractal.deep_sort import preprocessing
# from fractal.deep_sort import nn_matching
# from fractal.deep_sort.detection import Detection
# from fractal.deep_sort.tracker import Tracker
# from fractal.deep_sort.detection import Detection as ddet
# from fractal.deep_sort.feature_extractor import Extractor

# from fractal.superai.Nets.resnets import resnet18

class Inference:
    def __init__(self, cfg,tracker=None, extractor=None, iou = 0.5, conf = 0.5, heatmap=False, classifier=None):
        self.cfg = cfg
        self.model_name = cfg["model_stuff"]["model_name"]
        self.strides = cfg["arch"]["strides"]
        self.anchor_ratios = cfg["arch"]["anchor_ratios"]
        self.input_size = cfg["detection"]["input_size"]
        self.num_classes = cfg["data"]["num_classes"]
        self.num_anchors = len(cfg["arch"]["anchor_ratios"]) * len(cfg["arch"]["strides"])
        self.scale = cfg["arch"]["scale"]
        self.model = Darknet_yolo(num_classes=self.num_classes, anchors = self.num_anchors, scale = self.scale)
        self.model_weights = torch.load(cfg["detection"]["use_pretrained_weights_loc"])
        self.model.load_weights_manually(cfg["detection"]["use_pretrained_weights_loc"])

        self.encode = get_anchors(self.model_name)(self.strides,self.anchor_ratios,self.input_size)

        self.enc_anchors = self.encode.all_anchors_ft_scale.unsqueeze(0).clone()
        self.enc_strides = self.encode.all_anchors_strides.clone()
        self.heatmap=heatmap

        if self.cfg["device"]["use_gpu"]:
            self.model = self.model.cuda()
            self.enc_anchors = self.enc_anchors.cuda()
            self.enc_strides = self.enc_strides.cuda()

        self.labels_mapping = self.labels_mapper(cfg["data"]["labels_data"])
        self.iou = iou
        self.conf = conf
        self.tracker = tracker
        self.extractor = extractor
        # self.face_detector =  dlib.get_frontal_face_detector()
        self.face_detector = None
        self.track_list = {}
        self.idx = 0
        self.dwell_time = {}
        self.classifier = classifier
        self.face_detect = cfg["face_detection"]

    def infer_img(self, input_image_loc):
        """
        Extracts image from a given location, predicts it throgh the model and returns the image along with predicted bbox coordinates, 
        probabilities and labels
        """
        orig_img, img_tensor, shape = ListDataset.prep_image(input_image_loc, (self.input_size, self.input_size))
        if self.cfg["device"]["use_gpu"]:
            img_tensor = img_tensor.cuda()
        outputs = self.model.predict(img_tensor.cuda(), self.enc_anchors, self.enc_strides, iou = self.iou, conf=self.conf)
        # Filters to apply for which classes to use. If need all the classes, remove this bit.
        if not isinstance(outputs, int):
            outputs = outputs[outputs[:,7]==0]

        try:
            if not isinstance(outputs, int):
                h, w = shape 
                ow, oh = img_tensor.size()[2:]
                coord = outputs[:, 1:5]
                prob_scores = outputs[:, 6]
                labels  = outputs[:, 7]
                final_coord = reframe_bboxes_for_actual_image([h, w], [oh, ow], coord)
                final_coord = clamper(final_coord, [h, w])
                final_coord = final_coord.cpu().numpy()
                prob_scores = prob_scores.cpu().numpy()
                labels =  labels.cpu().numpy()
                return orig_img,final_coord, prob_scores, labels
            else:
                return orig_img,0,0,0
        except:
            return orig_img,0,0,0

    def infer_img_list(self, images_root, fnames, save_text = None, save_folder=None):
        """
        Infers a list of images, with their paths indicated with images_root and fnames, prints predicitons text file and 
        saves images with bounding boxes.
        """
        if not os.path.exists(save_folder):
           os.makedirs(save_folder)
        
        fnames = sorted(fnames)
        bar = progressbar.ProgressBar(maxval=len(fnames)).start()

        self.track_list ={}

        for num, img in enumerate(fnames):
            orig_img, coord, prob_scores, labels = self.infer_img(images_root+img)

            if isinstance(coord, int):
                if save_folder:
                    img_name = img.strip().split('/')[-1]
                    orig_img.save(save_folder+img_name)
            else:
                if save_folder:
                    img_name = img.strip().split('/')[-1]
                    frame = np.asarray(orig_img)
                    #coord=coord[np.where(prob_scores>0.9)]
                    #labels=labels[np.where(prob_scores>0.9)]
                    #prob_scores=prob_scores[np.where(prob_scores>0.9)]
                    if self.tracker:
                        frame = self.tracking(coord,frame, num, orig_img)
                        im = Image.fromarray(np.uint8(frame))
                        if self.heatmap:
                            im= self.heatmapper(im, coord)
                        im.save(save_folder + img_name)
                    else:
                        save_img_with_labels(orig_img, coord, prob_scores, labels,
                                                self.labels_mapping,
                                                save_loc=save_folder + img_name) 
            bar.update(num)
        bar.finish()       

    def labels_mapper(self,txtfile):
        """
        Reads the labels text file and returns labels against the respective class number in the dictionary.
        """
        labels = [i.split()[0] for i in open(txtfile, "r")]
        mapping = {i: labels.index(i) for i in labels}
        labels_mapping = {v:k for k, v in mapping.items()}
        return labels_mapping

    def tracking(self, coord, frame, frame_num, orig_img):
        boxs = np.empty_like(coord)
        boxs[:, 3] = coord[:,3] - coord[:,1]
        boxs[:, 2] = coord[:,2] - coord[:,0]
        boxs[:,1] = coord[:,1]
        boxs[:,0] = coord[:,0] 
        boxs = boxs.tolist()
        features = _get_features(frame, coord,self.extractor)
        detections = [Detection(bbox, 1.0, feature) for bbox, feature in zip(boxs, features)]               
        self.tracker.predict()
        self.tracker.update(detections)
        if self.face_detect:
            faces = self.face(frame)
        for track in self.tracker.tracks:
            if track.is_confirmed() and track.time_since_update >1 :
                continue                
            bbox = track.to_tlbr()
            if track.track_id in self.dwell_time:
                self.dwell_time[track.track_id][1] = frame_num
            else:
                self.dwell_time[track.track_id] = [frame_num, frame_num]
            # To classify betweek staff and customers
            if self.classifier:
                if track.track_id not in self.track_list :
                    self.track_list[track.track_id] = {}
                    self.track_list[track.track_id]["age"] = 0
                    self.track_list[track.track_id]["category"] = 0
                if self.track_list[track.track_id]["age"] < 25:
                    roi = orig_img.crop((int(bbox[0]), int(bbox[1]), int(bbox[2]), int(bbox[3])))
                    #roi.save("a.jpg")
                    category = self.classifier.predict_class(roi)
                    self.track_list[track.track_id]["age"] += 1
                    self.track_list[track.track_id]["category"] = category
                if self.track_list[track.track_id]["category"]==0:
                    cv2.putText(frame, str(track.track_id),(int((bbox[0]+bbox[2])/2), int((bbox[1]+bbox[3])/2)),0, 5e-3 * 250, (0,0,0),2)
                else:
                    cv2.putText(frame, str(track.track_id),(int((bbox[0]+bbox[2])/2), int((bbox[1]+bbox[3])/2)),0, 5e-3 * 250, (255,165,0),2)
            else:
                cv2.putText(frame, str(track.track_id),(int((bbox[0]+bbox[2])/2), int((bbox[1]+bbox[3])/2)),0, 5e-3 * 250, (255,165,0),2)
            """
            if track.track_id not in self.track_list:
                self.idx+=1
                print (self.idx)
                roi = frame[int(bbox[1]):int(bbox[3]), int(bbox[0]):int(bbox[2])].copy()
                name = "distinct/%05d.jpg" % (self.idx)
                self.track_list.append(track.track_id)
                cv2.imwrite(name,cv2.cvtColor(roi, cv2.COLOR_BGR2RGB))
            
            #cv2.rectangle(frame, (int(bbox[0]), int(bbox[1])), (int(bbox[2]), int(bbox[3])),(255,255,255), 2)

            #cv2.putText(frame, str(track.track_id),(int((bbox[0]+bbox[2])/2), int((bbox[1]+bbox[3])/2)),0, 5e-3 * 250, (0,0,0),2)
            """

        for det in detections:
            bbox = det.to_tlbr()
            #cv2.rectangle(frame,(int(bbox[0]), int(bbox[1])), (int(bbox[2]), int(bbox[3])),(255,0,0), 2)
        cv2.putText(frame, "Number of people in current frame: %d" %(len(detections)), (20,60), 0, 5e-3 * 250, (255,255,255), 2 )
        if self.face_detect:
            for bbox in faces:
                cv2.rectangle(frame,(int(bbox[0]), int(bbox[1])), (int(bbox[2]), int(bbox[3])),(0,0,255), 2)
        return frame
    
    def face(self,image):
        dets = self.face_detector(image,1)
        faces = []
        for i, d in enumerate(dets):
            xmin = min(d.left(),d.right())
            xmax = max(d.left(),d.right())
            ymin = min(d.top(), d.bottom())
            ymax = max(d.top(), d.bottom())
            faces.append([xmin,ymin,xmax,ymax])
        return faces

    def heatmapper(self, img, coord):
        points = []
        for crd in coord:
            ymin = min(int(crd[1]), int(crd[3]))
            xmin = min(int(crd[0]), int(crd[2]))
            ymax = max(int(crd[1]), int(crd[3]))
            xmax = max(int(crd[0]), int(crd[2]))
            X = np.arange(xmin - int(0.1*(xmax-xmin)),xmax - int(0.1*(xmax-xmin)) , 200)
            Y = np.arange(ymin,ymax, 200)
            pts = np.array(np.meshgrid(X, Y)).T.reshape(-1,2)
            pts = pts.tolist()
            points += pts
        heatmapper = Heatmapper(point_diameter=600)
        img = heatmapper.heatmap_on_img(points, img)
        return img
    
    def update_track(self, coord, frame, num, orig_img, save_folder):
        if self.tracker:
            frame = self.tracking(coord,frame, num, orig_img)
            im = Image.fromarray(np.uint8(frame))
            im.save(save_folder + img_name)
        else:
            save_img_with_labels(orig_img, coord, prob_scores, labels, labels_mapping, save_loc=save_folder + img_name)

def split_video_to_frames_ffmpeg(video_path, folder_root, name_template='img_%05d.png'):
    """
    splits a particular video into frams and saves the at mentioned folder_root
    """
    print("Splitting video to frames")
    os.system("mkdir -p {}".format(folder_root))
    cmd = 'ffmpeg -loglevel panic  -i {} -b:v 4500k {}/{}'.format(video_path, folder_root, name_template)
    start = time.time()
    os.system(cmd)
    end = time.time()
    print("{}: time: {}".format(os.path.basename(video_path), end - start))

def get_file_list(path):
    """
    gets list of file locations from a text file or folder
    """
    fnames = []
    if os.path.isfile(path):
        with open(path) as f:
            lines = f.readlines()

        for line in lines:
            splited = line.strip().split()
            fnames.append(splited[0])
    else:
        for file in os.listdir(path):
            fnames.append(file)
    return fnames

def save_text_preds(img_name, coord, prob_scores, labels, opener):
    """
    save the predictions of image in a text file. Very IMP change in change of coordinate system
    """
    adder = img_name
    coord = coord[:, [1,0, 3, 2]] ## xyxy to yxyx
    for xyxy, prob, label in zip(coord, prob_scores, labels):
        xyxy = list(xyxy)
        prob = str(float(prob))
        label = str(int(label))
        joiner = " " +" ".join([str(i) for i in xyxy])+ " "+label+ " "+prob
        adder +=joiner
    opener.write("{} \n".format(adder))

def fram_to_vid(pred_frames, destination, name_template):
    """
    recombines frames present in specified folder to a video.
    """
    print ("Recombining video frames to video")
    cmd = 'ffmpeg -loglevel panic -i {}{} -b:v 4500k -y {}'.format(pred_frames, name_template, destination)
    os.system(cmd)

def _get_features(ori_img, bbox_xyxy, extractor):
        features = []
        for box in bbox_xyxy:
            x1,y1,x2,y2 = box
            im = ori_img[int(y1):int(y2),int(x1):int(x2)]
            feature = extractor(im)[0]
            features.append(feature)
        
        if len(features) > 0:
            features = np.stack(features, axis=0)
        
        return features

class Classification:
    def __init__(self, cfg):
        self.model = resnet18(cfg["classification"]["num_classes"], pretrained=True)
        save = torch.load(cfg["classification"]["weights_loc"])
        self.use_gpu = cfg["device"]["use_gpu"]
        if self.use_gpu:
            self.model = self.model.cuda()
        self.model.load_state_dict(save["model_weights"])
        self.model = self.model.eval()
        
        
        self.image_transforms = []
        self.image_transforms.append(transforms.Resize((224,224)))
        self.image_transforms.append(transforms.ToTensor())
        self.image_transforms.append(transforms.Normalize((0.6357,  0.5425,  0.4426), (0.2740,  0.2873,  0.3129)))
        
        self.composed_transforms = transforms.Compose(self.image_transforms)
    
    def predict_class(self, inputs):
        img_tensor = self.composed_transforms(inputs)
        img_tensor = img_tensor.unsqueeze(0)
        if self.use_gpu:
            img_tensor = img_tensor.cuda()
        outputs = self.model(img_tensor)
        outputs = torch.nn.functional.softmax(outputs)
        outputs = outputs.squeeze()
        category = outputs.max(0)[1].item()
        return category


# vim: expandtab:ts=4:sw=4
from __future__ import absolute_import
import numpy as np
from . import kalman_filter
from . import linear_assignment
from . import iou_matching
from .track import Track


class Tracker:
    """
    This is the multi-target tracker.

    Parameters
    ----------
    metric : nn_matching.NearestNeighborDistanceMetric
        A distance metric for measurement-to-track association.
    max_age : int
        Maximum number of missed misses before a track is deleted.
    n_init : int
        Number of consecutive detections before the track is confirmed. The
        track state is set to `Deleted` if a miss occurs within the first
        `n_init` frames.

    Attributes
    ----------
    metric : nn_matching.NearestNeighborDistanceMetric
        The distance metric used for measurement to track association.
    max_age : int
        Maximum number of missed misses before a track is deleted.
    n_init : int
        Number of frames that a track remains in initialization phase.
    kf : kalman_filter.KalmanFilter
        A Kalman filter to filter target trajectories in image space.
    tracks : List[Track]
        The list of active tracks at the current time step.

    """

    def __init__(self, metric, max_iou_distance=0.7, max_age=30, n_init=3):
        self.metric = metric
        self.max_iou_distance = max_iou_distance
        self.max_age = max_age
        self.n_init = n_init

        self.kf = kalman_filter.KalmanFilter()
        self.tracks = []
        self._next_id = 1

    def predict(self):
        """Propagate track state distributions one time step forward.

        This function should be called once every time step, before `update`.
        """
        for track in self.tracks:
            track.predict(self.kf)

    def update(self, detections):
        """Perform measurement update and track management.

        Parameters
        ----------
        detections : List[deep_sort.detection.Detection]
            A list of detections at the current time step.

        """
        # Run matching cascade.
        matches, unmatched_tracks, unmatched_detections = \
            self._match(detections)

        # Update track set.
        for track_idx, detection_idx in matches:
            self.tracks[track_idx].update(
                self.kf, detections[detection_idx])
        for track_idx in unmatched_tracks:
            self.tracks[track_idx].mark_missed()
        for detection_idx in unmatched_detections:
            self._initiate_track(detections[detection_idx])
        self.tracks = [t for t in self.tracks if not t.is_deleted()]

        # Update distance metric.
        active_targets = [t.track_id for t in self.tracks if t.is_confirmed()]
        features, targets = [], []
        for track in self.tracks:
            if not track.is_confirmed():
                continue
            features += track.features
            targets += [track.track_id for _ in track.features]
            track.features = []
        self.metric.partial_fit(
            np.asarray(features), np.asarray(targets), active_targets)

    def _match(self, detections):

        def gated_metric(tracks, dets, track_indices, detection_indices):
            features = np.array([dets[i].feature for i in detection_indices])
            targets = np.array([tracks[i].track_id for i in track_indices])
            cost_matrix = self.metric.distance(features, targets)
            cost_matrix = linear_assignment.gate_cost_matrix(
                self.kf, cost_matrix, tracks, dets, track_indices,
                detection_indices)

            return cost_matrix

        # Split track set into confirmed and unconfirmed tracks.
        confirmed_tracks = [
            i for i, t in enumerate(self.tracks) if t.is_confirmed()]
        unconfirmed_tracks = [
            i for i, t in enumerate(self.tracks) if not t.is_confirmed()]

        # Associate confirmed tracks using appearance features.
        matches_a, unmatched_tracks_a, unmatched_detections = \
            linear_assignment.matching_cascade(
                gated_metric, self.metric.matching_threshold, self.max_age,
                self.tracks, detections, confirmed_tracks)

        # Associate remaining tracks together with unconfirmed tracks using IOU.
        iou_track_candidates = unconfirmed_tracks + [
            k for k in unmatched_tracks_a if
            self.tracks[k].time_since_update == 1]
        unmatched_tracks_a = [
            k for k in unmatched_tracks_a if
            self.tracks[k].time_since_update != 1]
        matches_b, unmatched_tracks_b, unmatched_detections = \
            linear_assignment.min_cost_matching(
                iou_matching.iou_cost, self.max_iou_distance, self.tracks,
                detections, iou_track_candidates, unmatched_detections)

        matches = matches_a + matches_b
        unmatched_tracks = list(set(unmatched_tracks_a + unmatched_tracks_b))
        return matches, unmatched_tracks, unmatched_detections

    def _initiate_track(self, detection):
        mean, covariance = self.kf.initiate(detection.to_xyah())
        self.tracks.append(Track(
            mean, covariance, self._next_id, self.n_init, self.max_age,
            detection.feature))
        self._next_id += 1


import time
import torch 
import torch.nn as nn
from torch.autograd import Variable
import numpy as np
import cv2 
from person_tracking.fractal.nets.darknet import Darknet_yolo
# from fractal.nets.darknet import Darknet_yolo
import random 
import argparse
import pickle as pkl
import torchvision.transforms as transforms
import json
import os
from person_tracking.fractal.anchors import get_anchors
from person_tracking.infer_utils import Inference, Classification, split_video_to_frames_ffmpeg, get_file_list, save_text_preds, fram_to_vid
# from fractal.anchors import get_anchors
# from infer_utils import Inference, Classification, split_video_to_frames_ffmpeg, get_file_list, save_text_preds, fram_to_vid

from person_tracking.fractal.deep_sort import preprocessing
from person_tracking.fractal.deep_sort import nn_matching
from person_tracking.fractal.deep_sort.detection import Detection
from person_tracking.fractal.deep_sort.tracker import Tracker
from person_tracking.fractal.deep_sort.detection import Detection as ddet
from person_tracking.fractal.deep_sort.feature_extractor import Extractor
# from fractal.deep_sort import preprocessing
# from fractal.deep_sort import nn_matching
# from fractal.deep_sort.detection import Detection
# from fractal.deep_sort.tracker import Tracker
# from fractal.deep_sort.detection import Detection as ddet
# from fractal.deep_sort.feature_extractor import Extractor

parser = argparse.ArgumentParser(description="Inference on image/list of images")
parser.add_argument("-cfg", "--cfg", default="cfgs/coco.json")
parser.add_argument("-vloc", "--video_location", help="Video location", default=None)
parser.add_argument("-sii", "--save_predictions_img", help="The folder where predicted image with bounding box will be saved ", default='../tmp_pred/')
parser.add_argument("-save", "--save_video_loc", default = "result.mp4")
args = parser.parse_args()

def Model(cfg_path: str):
    with open(cfg_path) as fp:
        cfg = json.load(fp)

    extractor = Extractor(cfg["tracking"]["weights_loc"], use_cuda=True)

    metric = nn_matching.NearestNeighborDistanceMetric(cfg["tracking"]["metric"], cfg["tracking"]["max_cosine_distance"], cfg["tracking"]["nn_budget"])
    tracker = Tracker(metric, max_iou_distance=cfg["tracking"]["max_iou_distance"], max_age=cfg["tracking"]["max_age"], n_init=cfg["tracking"]["n_init"])

    if cfg["classification"]["classify"]:
        classifier = Classification(cfg)
    else:
        classifier=None

    model = Inference(cfg, tracker, extractor, heatmap=cfg["heatmap"], classifier=classifier)
    return model

'''
model = Model(args.cfg)
start = time.time()

vid_root = args.video_location
fram_root = '../tmp/'
if not os.path.exists(fram_root):
    os.mkdir(fram_root)
# split_video_to_frames_ffmpeg(vid_root, fram_root, name_template='img_%05d.png')
fnames = get_file_list(fram_root)
model.infer_img_list(fram_root, fnames, save_folder=args.save_predictions_img)
print ("Dwell times for individual ID's of frames: ", model.dwell_time)

fram_to_vid(args.save_predictions_img, args.save_video_loc, name_template = 'img_%05d.png')
print (time.time()-start)
'''

'''
for vid in os.listdir('chunks'):
    vid_root = 'chunks/' + vid
    fram_root = './tmp/'
    if not os.path.exists(fram_root):
        os.mkdir(fram_root)
    split_video_to_frames_ffmpeg(vid_root, fram_root, name_template='img_%05d.png')
    fnames = get_file_list(fram_root)
    model.infer_img_list(fram_root, fnames, save_folder=args.save_predictions_img)

#fram_to_vid(args.save_predictions_img, args.save_video_loc, name_template = 'img_%05d.png')
print (time.time()-start)
'''


import torch
import torch.nn as nn
import torch.utils.model_zoo as model_zoo
import os
import sys

__all__ = ['InceptionResNetV2', 'inceptionresnetv2']

pretrained_settings = {
    'inceptionresnetv2': {
        'imagenet': {
            'url': 'http://data.lip6.fr/cadene/pretrainedmodels/inceptionresnetv2-520b38e4.pth',
            'input_space': 'RGB',
            'input_size': [3, 299, 299],
            'input_range': [0, 1],
            'mean': [0.5, 0.5, 0.5],
            'std': [0.5, 0.5, 0.5],
            'num_classes': 1000
        },
        'imagenet+background': {
            'url': 'http://data.lip6.fr/cadene/pretrainedmodels/inceptionresnetv2-520b38e4.pth',
            'input_space': 'RGB',
            'input_size': [3, 299, 299],
            'input_range': [0, 1],
            'mean': [0.5, 0.5, 0.5],
            'std': [0.5, 0.5, 0.5],
            'num_classes': 1001
        }
    }
}


class BasicConv2d(nn.Module):

    def __init__(self, in_planes, out_planes, kernel_size, stride, padding=0):
        super(BasicConv2d, self).__init__()
        self.conv = nn.Conv2d(in_planes, out_planes,
                              kernel_size=kernel_size, stride=stride,
                              padding=padding, bias=False) # verify bias false
        self.bn = nn.BatchNorm2d(out_planes,
                                 eps=0.001, # value found in tensorflow
                                 momentum=0.1, # default pytorch value
                                 affine=True)
        self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class Mixed_5b(nn.Module):

    def __init__(self):
        super(Mixed_5b, self).__init__()

        self.branch0 = BasicConv2d(192, 96, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(192, 48, kernel_size=1, stride=1),
            BasicConv2d(48, 64, kernel_size=5, stride=1, padding=2)
        ) 

        self.branch2 = nn.Sequential(
            BasicConv2d(192, 64, kernel_size=1, stride=1),
            BasicConv2d(64, 96, kernel_size=3, stride=1, padding=1),
            BasicConv2d(96, 96, kernel_size=3, stride=1, padding=1)
        )

        self.branch3 = nn.Sequential(
            nn.AvgPool2d(3, stride=1, padding=1, count_include_pad=False),
            BasicConv2d(192, 64, kernel_size=1, stride=1)
        )

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = self.branch3(x)
        out = torch.cat((x0, x1, x2, x3), 1)
        return out


class Block35(nn.Module):

    def __init__(self, scale=1.0):
        super(Block35, self).__init__()

        self.scale = scale

        self.branch0 = BasicConv2d(320, 32, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(320, 32, kernel_size=1, stride=1),
            BasicConv2d(32, 32, kernel_size=3, stride=1, padding=1)
        )

        self.branch2 = nn.Sequential(
            BasicConv2d(320, 32, kernel_size=1, stride=1),
            BasicConv2d(32, 48, kernel_size=3, stride=1, padding=1),
            BasicConv2d(48, 64, kernel_size=3, stride=1, padding=1)
        )

        self.conv2d = nn.Conv2d(128, 320, kernel_size=1, stride=1)
        self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        out = torch.cat((x0, x1, x2), 1)
        out = self.conv2d(out)
        out = out * self.scale + x
        out = self.relu(out)
        return out


class Mixed_6a(nn.Module):

    def __init__(self):
        super(Mixed_6a, self).__init__()
        
        self.branch0 = BasicConv2d(320, 384, kernel_size=3, stride=2)

        self.branch1 = nn.Sequential(
            BasicConv2d(320, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 256, kernel_size=3, stride=1, padding=1),
            BasicConv2d(256, 384, kernel_size=3, stride=2)
        )

        self.branch2 = nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        out = torch.cat((x0, x1, x2), 1)
        return out


class Block17(nn.Module):

    def __init__(self, scale=1.0):
        super(Block17, self).__init__()

        self.scale = scale

        self.branch0 = BasicConv2d(1088, 192, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(1088, 128, kernel_size=1, stride=1),
            BasicConv2d(128, 160, kernel_size=(1,7), stride=1, padding=(0,3)),
            BasicConv2d(160, 192, kernel_size=(7,1), stride=1, padding=(3,0))
        )

        self.conv2d = nn.Conv2d(384, 1088, kernel_size=1, stride=1)
        self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        out = torch.cat((x0, x1), 1)
        out = self.conv2d(out)
        out = out * self.scale + x
        out = self.relu(out)
        return out


class Mixed_7a(nn.Module):

    def __init__(self):
        super(Mixed_7a, self).__init__()
        
        self.branch0 = nn.Sequential(
            BasicConv2d(1088, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 384, kernel_size=3, stride=2)
        )

        self.branch1 = nn.Sequential(
            BasicConv2d(1088, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 288, kernel_size=3, stride=2)
        )

        self.branch2 = nn.Sequential(
            BasicConv2d(1088, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 288, kernel_size=3, stride=1, padding=1),
            BasicConv2d(288, 320, kernel_size=3, stride=2)
        )

        self.branch3 = nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = self.branch3(x)
        out = torch.cat((x0, x1, x2, x3), 1)
        return out


class Block8(nn.Module):

    def __init__(self, scale=1.0, noReLU=False):
        super(Block8, self).__init__()

        self.scale = scale
        self.noReLU = noReLU

        self.branch0 = BasicConv2d(2080, 192, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(2080, 192, kernel_size=1, stride=1),
            BasicConv2d(192, 224, kernel_size=(1,3), stride=1, padding=(0,1)),
            BasicConv2d(224, 256, kernel_size=(3,1), stride=1, padding=(1,0))
        )

        self.conv2d = nn.Conv2d(448, 2080, kernel_size=1, stride=1)
        if not self.noReLU:
            self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        out = torch.cat((x0, x1), 1)
        out = self.conv2d(out)
        out = out * self.scale + x
        if not self.noReLU:
            out = self.relu(out)
        return out


class InceptionResNetV2(nn.Module):

    def __init__(self, num_classes=1001):
        super(InceptionResNetV2, self).__init__()
        # Special attributs
        self.input_space = None
        self.input_size = (299, 299, 3)
        self.mean = None
        self.std = None
        # Modules
        self.conv2d_1a = BasicConv2d(3, 32, kernel_size=3, stride=2)
        self.conv2d_2a = BasicConv2d(32, 32, kernel_size=3, stride=1)
        self.conv2d_2b = BasicConv2d(32, 64, kernel_size=3, stride=1, padding=1)
        self.maxpool_3a = nn.MaxPool2d(3, stride=2)
        self.conv2d_3b = BasicConv2d(64, 80, kernel_size=1, stride=1)
        self.conv2d_4a = BasicConv2d(80, 192, kernel_size=3, stride=1)
        self.maxpool_5a = nn.MaxPool2d(3, stride=2)
        self.mixed_5b = Mixed_5b()
        self.repeat = nn.Sequential(
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17),
            Block35(scale=0.17)
        )
        self.mixed_6a = Mixed_6a()
        self.repeat_1 = nn.Sequential(
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10),
            Block17(scale=0.10)
        )
        self.mixed_7a = Mixed_7a()
        self.repeat_2 = nn.Sequential(
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20),
            Block8(scale=0.20)
        )
        self.block8 = Block8(noReLU=True)
        self.conv2d_7b = BasicConv2d(2080, 1536, kernel_size=1, stride=1)
        self.avgpool_1a = nn.AvgPool2d(8, count_include_pad=False)
        self.last_linear = nn.Linear(1536, num_classes)

    def features(self, input):
        x = self.conv2d_1a(input)
        x = self.conv2d_2a(x)
        x = self.conv2d_2b(x)
        x = self.maxpool_3a(x)
        x = self.conv2d_3b(x)
        x = self.conv2d_4a(x)
        x = self.maxpool_5a(x)
        x = self.mixed_5b(x)
        x = self.repeat(x)
        x = self.mixed_6a(x)
        x = self.repeat_1(x)
        x = self.mixed_7a(x)
        x = self.repeat_2(x)
        x = self.block8(x)
        x = self.conv2d_7b(x)
        return x

    def logits(self, features):
        x = self.avgpool_1a(features)
        x = x.view(x.size(0), -1)
        x = self.last_linear(x) 
        return x

    def forward(self, input):
        x = self.features(input)
        x = self.logits(x)
        return x

def inceptionresnetv2(num_classes=1000, pretrained='imagenet'):
    r"""InceptionResNetV2 model architecture from the
    `"InceptionV4, Inception-ResNet..." <https://arxiv.org/abs/1602.07261>`_ paper.
    """
    if pretrained == "imagenet":
        settings = pretrained_settings['inceptionresnetv2'][pretrained]
        model = InceptionResNetV2(num_classes=1001)
        model.load_state_dict(model_zoo.load_url(settings['url']))
        
        new_last_linear = nn.Linear(1536, num_classes)
        model.last_linear = new_last_linear
        
    else:
        model = InceptionResNetV2(num_classes=num_classes)
    return model